# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The microversion_parse middlware is tests using gabbi to run real
# http requests through it. To do that, we need a simple WSGI
# application running under wsgi-intercept (handled by gabbi).

import os

from gabbi import driver
import webob

from microversion_parse import middleware


TESTS_DIR = 'gabbits'
SERVICE_TYPE = 'cats'
VERSIONS = [
    '1.0',  # initial version
    '1.1',  # now with kittens
    '1.2',  # added breeds
]


class SimpleWSGI(object):
    """A WSGI application that can be contiained within a middlware."""

    def __call__(self, environ, start_response):
        path_info = environ['PATH_INFO']
        if path_info == '/good':
            start_response('200 OK', [('content-type', 'text/plain')])
            return [b'good']

        raise webob.exc.HTTPNotFound('%s not found' % path_info)


def app():
    app = middleware.MicroversionMiddleware(
            SimpleWSGI(), SERVICE_TYPE, VERSIONS)
    return app


def load_tests(loader, tests, pattern):
    """Provide a TestSuite to the discovery process."""
    test_dir = os.path.join(os.path.dirname(__file__), TESTS_DIR)
    return driver.build_tests(
        test_dir, loader, test_loader_name=__name__, intercept=app)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

import microversion_parse


class TestHeadersFromWSGIEnviron(testtools.TestCase):

    def test_empty_environ(self):
        environ = {}
        expected = {}
        self.assertEqual(
            expected,
            microversion_parse.headers_from_wsgi_environ(environ))

    def test_non_empty_no_headers(self):
        environ = {'PATH_INFO': '/foo/bar'}
        expected = {}
        found_headers = microversion_parse.headers_from_wsgi_environ(environ)
        self.assertEqual(expected, found_headers)

    def test_headers(self):
        environ = {'PATH_INFO': '/foo/bar',
                   'HTTP_OPENSTACK_API_VERSION': 'placement 2.1',
                   'HTTP_CONTENT_TYPE': 'application/json'}
        expected = {'HTTP_OPENSTACK_API_VERSION': 'placement 2.1',
                    'HTTP_CONTENT_TYPE': 'application/json'}
        found_headers = microversion_parse.headers_from_wsgi_environ(environ)
        self.assertEqual(expected, found_headers)

    def test_get_version_from_environ(self):
        environ = {'PATH_INFO': '/foo/bar',
                   'HTTP_OPENSTACK_API_VERSION': 'placement 2.1',
                   'HTTP_CONTENT_TYPE': 'application/json'}
        expected_version = '2.1'
        headers = microversion_parse.headers_from_wsgi_environ(environ)
        version = microversion_parse.get_version(headers, 'placement')
        self.assertEqual(expected_version, version)

    def test_get_version_from_environ_legacy(self):
        environ = {'PATH_INFO': '/foo/bar',
                   'HTTP_X_OPENSTACK_PLACEMENT_API_VERSION': '2.1',
                   'HTTP_CONTENT_TYPE': 'application/json'}
        expected_version = '2.1'
        headers = microversion_parse.headers_from_wsgi_environ(environ)
        version = microversion_parse.get_version(
            headers, 'placement',
            legacy_headers=['x-openstack-placement-api-version'])
        self.assertEqual(expected_version, version)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import testtools

import microversion_parse


class TestVersion(testtools.TestCase):

    def setUp(self):
        super(TestVersion, self).setUp()
        self.version = microversion_parse.Version(1, 5)

    def test_version_is_tuple(self):
        self.assertEqual((1, 5), self.version)

    def test_version_stringifies(self):
        self.assertEqual('1.5', str(self.version))

    def test_version_matches(self):
        max_version = microversion_parse.Version(1, 20)
        min_version = microversion_parse.Version(1, 3)

        self.assertTrue(self.version.matches(min_version, max_version))
        self.assertFalse(self.version.matches(max_version, min_version))

    def test_version_matches_inclusive(self):
        max_version = microversion_parse.Version(1, 5)
        min_version = microversion_parse.Version(1, 5)

        self.assertTrue(self.version.matches(min_version, max_version))

    def test_version_matches_no_extremes(self):
        """If no extremes are present, never match."""
        self.assertFalse(self.version.matches())

    def test_version_zero_can_match(self):
        """If a version is '0.0' we want to it be able to match."""
        version = microversion_parse.Version(0, 0)
        min_version = microversion_parse.Version(0, 0)
        max_version = microversion_parse.Version(0, 0)
        version.min_version = min_version
        version.max_version = max_version

        self.assertTrue(version.matches())

    def test_version_zero_no_defaults(self):
        """Any version, even 0.0, should never match without a min
        and max being set.
        """
        version = microversion_parse.Version(0, 0)

        self.assertFalse(version.matches())

    def test_version_init_failure(self):
        self.assertRaises(TypeError, microversion_parse.Version, 1, 2, 3)


class TestParseVersionString(testtools.TestCase):

    def test_good_version(self):
        version = microversion_parse.parse_version_string('1.1')
        self.assertEqual((1, 1), version)
        self.assertEqual(microversion_parse.Version(1, 1), version)

    def test_adapt_whitespace(self):
        version = microversion_parse.parse_version_string(' 1.1 ')
        self.assertEqual((1, 1), version)
        self.assertEqual(microversion_parse.Version(1, 1), version)

    def test_non_numeric(self):
        self.assertRaises(TypeError,
                          microversion_parse.parse_version_string,
                          'hello')

    def test_mixed_alphanumeric(self):
        self.assertRaises(TypeError,
                          microversion_parse.parse_version_string,
                          '1.a')

    def test_too_many_numeric(self):
        self.assertRaises(TypeError,
                          microversion_parse.parse_version_string,
                          '1.1.1')

    def test_not_string(self):
        self.assertRaises(TypeError,
                          microversion_parse.parse_version_string,
                          1.1)


class TestExtractVersion(testtools.TestCase):

    def setUp(self):
        super(TestExtractVersion, self).setUp()
        self.headers = [
            ('OpenStack-API-Version', 'service1 1.2'),
            ('OpenStack-API-Version', 'service2 1.5'),
            ('OpenStack-API-Version', 'service3 latest'),
            ('OpenStack-API-Version', 'service4 2.5'),
        ]
        self.version_list = ['1.1', '1.2', '1.3', '1.4',
                             '2.1', '2.2', '2.3', '2.4']

    def test_simple_extract(self):
        version = microversion_parse.extract_version(
            self.headers, 'service1', self.version_list)
        self.assertEqual((1, 2), version)

    def test_default_min(self):
        version = microversion_parse.extract_version(
            self.headers, 'notlisted', self.version_list)
        self.assertEqual((1, 1), version)

    def test_latest(self):
        version = microversion_parse.extract_version(
            self.headers, 'service3', self.version_list)
        self.assertEqual((2, 4), version)

    def test_min_max_extract(self):
        version = microversion_parse.extract_version(
            self.headers, 'service1', self.version_list)

        # below min
        self.assertFalse(version.matches((1, 3)))
        # at min
        self.assertTrue(version.matches((1, 2)))
        # within extremes
        self.assertTrue(version.matches())
        # explicit max
        self.assertTrue(version.matches(max_version=(2, 3)))
        # explicit min
        self.assertFalse(version.matches(min_version=(2, 3)))
        # explicit both
        self.assertTrue(version.matches(min_version=(0, 3),
                                        max_version=(1, 5)))

    def test_version_disabled(self):
        self.assertRaises(ValueError, microversion_parse.extract_version,
                          self.headers, 'service2', self.version_list)

    def test_version_out_of_range(self):
        self.assertRaises(ValueError, microversion_parse.extract_version,
                          self.headers, 'service4', self.version_list)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""WSGI middleware for getting microversion info."""

import webob
import webob.dec

import microversion_parse


class MicroversionMiddleware(object):
    """WSGI middleware for getting microversion info.

    The application will get a WSGI environ with a
    'SERVICE_TYPE.microversion' key that has a value of the microversion
    found at an 'openstack-api-version' header that matches SERVICE_TYPE. If
    no header is found, the minimum microversion will be set. If the
    special keyword 'latest' is used, the maximum microversion will be
    set.

    If the requested microversion is not available a 406 response is
    returned.

    If there is an error parsing a provided header, a 400 response is
    returned.

    Otherwise the application is called.
    """

    def __init__(self, application, service_type, versions,
                 json_error_formatter=None):
        """Create the WSGI middleware.

        :param application: The application hosting the service.
        :param service_type: The service type (entry in keystone catalog)
                             of the application.
        :param versions: An ordered list of legitimate versions for the
                         application.
        :param json_error_formatter: A Webob exception error formatter.
                                     See Webob for details.
        """
        self.application = application
        self.service_type = service_type
        self.microversion_environ = '%s.microversion' % service_type
        self.versions = versions
        self.json_error_formatter = json_error_formatter

    @webob.dec.wsgify
    def __call__(self, req):
        try:
            microversion = microversion_parse.extract_version(
                req.headers, self.service_type, self.versions)
        # TODO(cdent): These error response are not formatted according to
        # api-sig guidelines, unless a json_error_formatter is provided
        # that can do it. For an example, see the placement service.
        except ValueError as exc:
            raise webob.exc.HTTPNotAcceptable(
                ('Invalid microversion: %(error)s') % {'error': exc},
                json_formatter=self.json_error_formatter)
        except TypeError as exc:
            raise webob.exc.HTTPBadRequest(
                ('Invalid microversion: %(error)s') % {'error': exc},
                json_formatter=self.json_error_formatter)

        req.environ[self.microversion_environ] = microversion
        microversion_header = '%s %s' % (self.service_type, microversion)
        standard_header = microversion_parse.STANDARD_HEADER

        try:
            response = req.get_response(self.application)
        except webob.exc.HTTPError as exc:
            # If there was an HTTPError in the application we still need
            # to send the microversion header, so add the header and
            # re-raise the exception.
            exc.headers.add(standard_header, microversion_header)
            raise exc

        response.headers.add(standard_header, microversion_header)
        response.headers.add('vary', standard_header)
        return response



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

from webob import headers as wb_headers

import microversion_parse


class TestWebobHeaders(testtools.TestCase):
    """Webob uses a dict-like header which is not actually a dict."""

    def test_simple_headers(self):
        headers = wb_headers.EnvironHeaders({
            'HTTP_HEADER_ONE': 'alpha',
            'HTTP_HEADER_TWO': 'beta',
            'HTTP_HEADER_THREE': 'gamma',
        })

        folded_headers = microversion_parse.fold_headers(headers)
        self.assertEqual(3, len(folded_headers))
        self.assertEqual(set(['header-one', 'header-three', 'header-two']),
                         set(folded_headers.keys()))
        self.assertEqual('gamma', folded_headers['header-three'])

    def test_simple_match(self):
        headers = wb_headers.EnvironHeaders({
            'HTTP_HEADER_ONE': 'alpha',
            'HTTP_OPENSTACK_API_VERSION': 'compute 2.1',
            'HTTP_HEADER_TWO': 'beta',
        })
        version = microversion_parse.check_standard_header(headers, 'compute')
        self.assertEqual('2.1', version)

    def test_match_multiple_services(self):
        headers = wb_headers.EnvironHeaders({
            'HTTP_HEADER_ONE': 'alpha',
            'HTTP_OPENSTACK_API_VERSION':
            'network 5.9 ,compute 2.1,telemetry 7.8',
            'HTTP_HEADER_TWO': 'beta',
        })
        version = microversion_parse.check_standard_header(
            headers, 'compute')
        self.assertEqual('2.1', version)
        version = microversion_parse.check_standard_header(
            headers, 'telemetry')
        self.assertEqual('7.8', version)

    def test_legacy_headers_straight(self):
        headers = wb_headers.EnvironHeaders({
            'HTTP_HEADER_ONE': 'alpha',
            'HTTP_X_OPENSTACK_NOVA_API_VERSION': ' 2.1 ',
            'HTTP_HEADER_TWO': 'beta',
        })
        version = microversion_parse.get_version(
            headers, service_type='compute',
            legacy_headers=['x-openstack-nova-api-version'])
        self.assertEqual('2.1', version)

    def test_legacy_headers_folded(self):
        headers = wb_headers.EnvironHeaders({
            'HTTP_HEADER_ONE': 'alpha',
            'HTTP_X_OPENSTACK_NOVA_API_VERSION': ' 2.1, 9.2 ',
            'HTTP_HEADER_TWO': 'beta',
        })
        version = microversion_parse.get_version(
            headers, service_type='compute',
            legacy_headers=['x-openstack-nova-api-version'])
        self.assertEqual('9.2', version)
